module Hv
  class Schema

    def initialize(schema={}, &block)
      @schema = schema
      @custom_validators = {}
      @custom_validators_keys = []
      @errors_processor = Hv::ErrorsProcessor
      @i18n_errors = Hv::ErrorsProcessor.i18n_errors
      self.instance_eval(&block) if block_given?
    end

    def result
      @result
    end

    def set_schema(schema)
      @schema = schema
    end

    def validates?(input)
      @skip_paths = []
      @result = []
      validate_schema(@schema, input, [])
      run_custom_validators(input)
      @result.empty?
    end

    def errors(errors_processor=nil, i18n_errors=nil)
      i18n_errors = i18n_errors || @i18n_errors
      errors_processor = errors_processor || @errors_processor
      errors_processor.call(@result, i18n_errors)
    end

    def validate(custom_validator_name, &block)
      @custom_validators[custom_validator_name] = block
    end

    def validator(key, &block)
      @custom_validators_keys << key
      define_singleton_method "validate_#{key}", block
    end

    def set_error_processor(error_processor)
      @errors_processor = error_processor
    end

    def set_i18n_errors(hash)
      @i18n_errors = hash
    end

    private

    def validator_key?(key)
      allowed_validators.include?(key) ||
      @custom_validators_keys.include?(key)
    end

    def allowed_validators
      [:optional?, :format?, :min_size?, :max_size?, :size?, :lt?, :gt?, :types?, :type?, :enum?, :proc?, :array?, :empty?]
    end

    def parent_is_optional?(path)
      @skip_paths.any? { |parent_path|
        (parent_path - path).empty?
      }
    end

    def run_custom_validators(input)
      @custom_validators.each do |name, blk|
        blk.call(input)
      end
    end

    def add_error(path, key, _spec, value)
      @result << [path, key, _spec, value]
    end

    def validate_schema(spec, value, path=nil)
      spec.each do |key, _spec|
        if !validator_key?(key)
          _value = value[key] rescue nil
          validate_schema(_spec, _value, [path, key].flatten)
        else
          match = if value.nil? && optional_attr?(path)
            @skip_paths << path
            true
          else
            next if key == :optional?
            next if parent_is_optional?(path)
            send("validate_#{key}",_spec, value, path)
          end
          add_error(path, key, _spec, value) unless match
        end
      end
    end

    def optional_attr?(path=[])
      if path.empty?
        @schema[:optional?]
      else
        (@schema.dig(*path) || {})[:optional?]
      end
    end

    def validate_enum?(spec, value, path)
      spec.include?(value)
    end

    def validate_types?(spec, value, path)
      raise "Argument for :types? must be an array of types" unless spec.is_a? Array
      spec.any? { |_type| value.is_a?(_type) }
    end


    def validate_type?(spec, value, path)
      value.is_a?(spec)
    end

    def validate_size(spec, value, oper)
      case value
      when Enumerable, String
        value.size.send(oper, spec)
      when Numeric
        value.send(oper, spec)
      end
    end

    def validate_min_size?(spec, value, path)
      validate_size(spec, value, '>=')
    end

    def validate_max_size?(spec, value, path)
      validate_size(spec, value, '<=')
    end

    def validate_size?(spec, value, path)
      validate_size(spec, value, '==')
    end

    def validate_lt?(spec, value, path)
      validate_size(spec, value, '<')
    end

    def validate_gt?(spec, value, path)
      validate_size(spec, value, '>')
    end

    def validate_format?(spec, value, path)
      if value.respond_to? :match?
        value.match?(spec)
      else
        false
      end
    end

    def validate_array?(spec, values, path=nil)
      return false unless values.is_a?(Array)
      values.to_enum.with_index.all? {|value, index|
        validate_schema(spec, value, [path, index].flatten)
      }
    end

  end
end
